"""Super simple checkpoints using numpy."""

import datetime
import os
from typing import Optional
import zipfile
import pickle 

import jax
import jax.numpy as jnp
import numpy as np

def find_last_checkpoint(ckpt_path: Optional[str] = None) -> Optional[str]:
  """Finds most recent valid checkpoint in a directory.
  Args:
    ckpt_path: Directory containing checkpoints.
  Returns:
    Last QMC checkpoint (ordered by sorting all checkpoints by name in reverse)
    or None if no valid checkpoint is found or ckpt_path is not given or doesn't
    exist. A checkpoint is regarded as not valid if it cannot be read
    successfully using np.load.
  """
  if ckpt_path and os.path.exists(ckpt_path):
    files = [f for f in os.listdir(ckpt_path) if ('kejax_ckpt_' in f) and ('npz' in f)]
    # Handle case where last checkpoint is corrupt/empty.
    for file in sorted(files, reverse=True):
      fname = os.path.join(ckpt_path, file)
      with open(fname, 'rb') as f:
        try:
          np.load(f, allow_pickle=True)
          return fname
        except (OSError, EOFError, zipfile.BadZipFile):
          print('Error loading checkpoint. Trying next checkpoint...',
                       fname)
  return None


def create_save_path(save_path: Optional[str]) -> str:
  """Creates the directory for saving checkpoints, if it doesn't exist.
  Args:
    save_path: directory to use. If false, create a directory in the working
      directory based upon the current time.
  Returns:
    Path to save checkpoints to.
  """
  timestamp = datetime.datetime.now().strftime('%Y_%m_%d_%H:%M:%S')
  default_save_path = os.path.join(os.getcwd(), f'ferminet_{timestamp}')
  ckpt_save_path = save_path or default_save_path
  if ckpt_save_path and not os.path.isdir(ckpt_save_path):
    os.makedirs(ckpt_save_path)
  return ckpt_save_path


def get_restore_path(restore_path: Optional[str] = None) -> Optional[str]:
  """Gets the path containing checkpoints from a previous calculation.
  Args:
    restore_path: path to checkpoints.
  Returns:
    The path or None if restore_path is falsy.
  """
  if restore_path:
    ckpt_restore_path = restore_path
  else:
    ckpt_restore_path = None
  return ckpt_restore_path


def save(save_path: str, t: int, params) -> str:
  """Saves checkpoint information to a npz file.
  Args:
    save_path: path to directory to save checkpoint to. The checkpoint file is
      save_path/qmcjax_ckpt_$t.npz, where $t is the number of completed
      iterations.
    t: number of completed iterations.
    params: pytree of network parameters.
  Returns:
    path to checkpoint file.
  """
  ckpt_filename = os.path.join(save_path, f'kejax_ckpt_{t:06d}.npz')
  print('Saving checkpoint', ckpt_filename)
  with open(ckpt_filename, 'wb') as f:
    np.savez(
        f,
        t=t)

  with open(ckpt_filename.replace('npz', 'pkl'), 'wb') as f:
    pickle.dump(params, f) 
  return ckpt_filename

def restore(restore_filename: str):
  """Restores data saved in a checkpoint.
  Args:
    restore_filename: filename containing checkpoint.
  Returns:
    (t, params) tuple, where
    t: number of completed iterations.
    params: pytree of network parameters.
  """
  print('Loading checkpoint', restore_filename)
  with open(restore_filename, 'rb') as f:
    ckpt_data = np.load(f, allow_pickle=True)
    # Retrieve data from npz file. Non-array variables need to be converted back
    # to natives types using .tolist().
    t = ckpt_data['t'].tolist() + 1  # Return the iterations completed.
  with open(restore_filename.replace('npz', 'pkl'), 'rb') as f:
    params = pickle.load(f)
  return t, params

