import jax
import jax.numpy as jnp

'''
cf https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.108.253002
'''
def make_training_data(N, G):

    xmesh = jnp.linspace(0, 1, G, endpoint=True)   
    dx = xmesh[1] - xmesh[0]

    K  = -0.5/dx**2 * (jnp.diag(-2 * jnp.ones(G))
            + jnp.diag(jnp.ones(G - 1), k=1)
            + jnp.diag(jnp.ones(G - 1), k=-1))

    def solver(key):
        #build potential 
        v = jnp.zeros_like(xmesh)
        for _ in range(3):
            r = jax.random.uniform(key, (3,))
            a = 9*r[0] + 1
            b = 0.2*r[1] + 0.4
            c = 0.07*r[2] + 0.03 
            v = v - a * jnp.exp(-(xmesh-b)**2/(2*c**2))
    
        #and then solve it
        w, psi = jnp.linalg.eigh(K + jnp.diag(v)) 
        rho = jnp.sum(psi[:, :N]**2, axis=1)
        return v, rho, jnp.sum(w[:N]) - jnp.sum(rho * v), w[N-1]

    def get_batch(key, batchsize):
        subkeys = jax.random.split(key, batchsize)
        return jax.vmap(solver)(subkeys) 

    return dx, xmesh, K, get_batch

if __name__=='__main__':
    import jax.scipy.optimize
    import matplotlib.pyplot as plt
    N = 1
    G = 500
    batchsize = 10
    dx, xmesh, K, get_batch = make_training_data(N, G)
    
    key = jax.random.PRNGKey(42)
    vs, rhos, Es, mus = get_batch(key, batchsize)

    xs = jnp.concatenate([jnp.sqrt(rhos), jnp.expand_dims(mus, axis=1)], axis=1)
    x0 = jnp.mean(xs, axis=0)
    print (xs.shape)

    print ('mus', mus)

    def total_energy_functional(x, v):
        phi, mu = x[:-1], x[-1]
        rho = phi**2
        #t = (jnp.gradient(rho)/dx)**2/rho/8
        return jnp.sum(rho * (v-mu)) + jnp.dot(jnp.dot(phi, K), phi)  + mu * N 

    print ('Es using functional', jax.vmap(total_energy_functional, (0, 0))(xs, vs)) 

    def euler_solver(x0, v):
        f = lambda x: total_energy_functional(x, v)
        res = jax.scipy.optimize.minimize(f, x0, method='BFGS')
        return res.x, res.fun

    def predict(vs):
        xs_star, Es_star = jax.vmap(euler_solver, (0, 0))(xs, vs)
        phis_star, mus_star = xs_star[:,:-1] , xs_star[:, -1]
        rhos_star = phis_star**2
        return rhos_star, Es_star, mus_star

    predicted_rhos, predicted_Es, predicted_mus = predict(vs)
    
    print ('predicted N', jnp.sum(predicted_rhos, axis=1))
    print ('predicted Es', predicted_Es)
    print ('predicted mus', predicted_mus)

    for i in range(batchsize):
        plt.plot(xmesh, predicted_rhos[i, :], label=r'$\rho$')
        plt.plot(xmesh, rhos[i, :], label='true')
        plt.legend()
        plt.show()
