import jax
from jax.config import config   
config.update("jax_enable_x64", True)
#config.update('jax_disable_jit', True)

import jax.numpy as jnp
import io, time
import matplotlib.pyplot as plt

from data import make_training_data
from model import make_network
from loss import make_loss
from args import args
import checkpoint

ckpt_restore_path = checkpoint.get_restore_path(args.restore_path)
ckpt_restore_filename = checkpoint.find_last_checkpoint(ckpt_restore_path)

dx, xmesh, K, get_batch = make_training_data(args.N, args.G) 
network = make_network(args.depth, args.channels, args.kernelsize)

key = jax.random.PRNGKey(42)
params = network.init(key, xmesh)

if ckpt_restore_filename:
    t_init, params = checkpoint.restore(ckpt_restore_filename)
else:
    print('No checkpoint found')

vs, rhos, Es, mus = get_batch(key, args.batchsize)
ts = jax.vmap(network.apply, (None, None, 0))(params, None, rhos)
print(ts.shape, jnp.sum(rhos* ts, axis=1).shape)

rho0 = jnp.mean(jnp.sqrt(rhos), axis=0)
density_loss, energy_loss, potential_loss, predict_fn, kinetic_energy_fn = make_loss(network, K, rho0, args.N, dx, args.max_iter, args.epsilon)

print ('energy_loss:', energy_loss(params, vs, rhos, Es, mus))
print ('potential_loss:', potential_loss(params, vs, rhos, Es, mus))
print ('density_loss:', density_loss(params, vs, rhos, Es, mus))

predicted_rhos, converge = predict_fn(params, vs)

if not converge:
    print ('!!! Euler solver did not converge in %g steps to reach %g accuracy'%(args.max_iter, args.epsilon))

for i in range(args.batchsize):
    plt.subplot(121)
    plt.plot(xmesh, predicted_rhos[i, :], label=r'$\rho$')
    plt.plot(xmesh, rhos[i, :], label='true')
    plt.legend()

    plt.subplot(122)
    plt.plot(xmesh, jax.grad(kinetic_energy_fn)(rhos[i, :], params), label=r'$\frac{\delta T}{\delta \rho}$')
    plt.plot(xmesh, mus[i]-vs[i, :], label=r'$\mu-v$')

    plt.legend()
    plt.show()
