import jax 
import jax.numpy as jnp
import jax.scipy.optimize
import numpy as np

import jaxopt

def make_loss(network, K, x0, N, dx, max_iter, epsilon):

    def kinetic_energy(rho, params):
        phi = jnp.sqrt(rho)
        t = jnp.sum(network.apply(params, None, rho)*rho)
        return jnp.dot(jnp.dot(phi, K), phi) + t
 
    def ofdft_solver(rho, v, params, lmb=4.0, alpha=0.5):
        #cf. https://journals.aps.org/prresearch/abstract/10.1103/PhysRevResearch.3.033198 eq.(16) for lmb 
        #cf. https://journals.aps.org/prl/abstract/10.1103/PhysRevLett.126.036401 eq.(S5) for alpha

        phi = jnp.sqrt(rho)
        t = lambda rho: jnp.sum(network.apply(params, None, rho)*rho)
        vs = (v + jax.grad(t)(rho) + (1.0-lmb) * jnp.dot(K, phi)/phi)/lmb
        _, psi = jnp.linalg.eigh(K + jnp.diag(vs)) 

        return rho + alpha*((psi[:,0]**2)*N - rho) 

    @jaxopt.implicit_diff.custom_fixed_point(ofdft_solver, has_aux=True)
    def euler_solver(x0, v, params):
        
        start, end = 0.5, 0.1 # decay mix parameter from start to end in max_iter steps
        base = np.exp(np.log(end/start)/max_iter)
        _alpha = lambda i: start*(base**i)

        def _body_fn(carry):
            i, _, x = carry 
            return i+1, x, ofdft_solver(x,v,params,alpha=_alpha(i))
        
        def _cond_fn(carry):
            i, x_prev, x = carry
            return (i< max_iter) & (jnp.linalg.norm(x_prev-x) > epsilon)
        
        init_carry = (0, x0, ofdft_solver(x0,v,params,alpha=_alpha(0)))
        n_iter, _, x_star = jax.lax.while_loop(_cond_fn, _body_fn, init_carry)
        return x_star, (n_iter < max_iter)

    def density_loss(params, vs, rhos, Es, mus):
        del Es, mus
        rhos_star, _ = predict(params, vs)
        return jnp.mean(jnp.sum(((rhos_star - rhos)/dx)**2*dx, axis=1)) 

    def energy_loss(params, vs, rhos, Es, mus):
        del vs, mus
        Ts = jax.vmap(kinetic_energy, (0, None))(rhos, params)
        return jnp.mean((Ts - Es)**2)

    def potential_loss(params, vs, rhos, Es, mus):
        del Es
        kfd = jax.vmap(jax.grad(kinetic_energy), (0, None))(rhos, params)
        return jnp.mean(jnp.sum((kfd + vs - jnp.expand_dims(mus, axis=1))**2*dx, axis=1)) 
    
    @jax.jit
    def predict(params, vs):
        rhos_star, converge = jax.vmap(euler_solver, (None, 0, None))(x0, vs, params)
        return rhos_star, jnp.alltrue(converge)

    return density_loss, energy_loss, potential_loss, predict, kinetic_energy

