import jax
from jax.config import config   
config.update("jax_enable_x64", True)

import jax.numpy as jnp
from jax.experimental import optimizers
from tqdm import trange
import io, time

from data import make_training_data
from model import make_network
from loss import make_loss
from args import args, save_path
import checkpoint

print ('Save path:', save_path)
ckpt_save_path = checkpoint.create_save_path(save_path)
ckpt_restore_path = checkpoint.get_restore_path(args.restore_path)

ckpt_restore_filename = (
     checkpoint.find_last_checkpoint(ckpt_restore_path) or
     checkpoint.find_last_checkpoint(ckpt_save_path)
     )

dx, xmesh, K, get_batch = make_training_data(args.N, args.G) 
network = make_network(args.depth, args.channels, args.kernelsize)

key = jax.random.PRNGKey(42)
params = network.init(key, xmesh)

if ckpt_restore_filename:
    t_init, params = checkpoint.restore(ckpt_restore_filename)
else:
    print('No checkpoint found. Training new model.')
    t_init = 0

vs, rhos, Es, mus = get_batch(key, args.batchsize)
ts = jax.vmap(network.apply, (None, None, 0))(params, None, rhos)
print(ts.shape, jnp.sum(rhos* ts, axis=1).shape)

rho0 = jnp.mean(jnp.sqrt(rhos), axis=0)
density_loss, energy_loss, potential_loss, predict_fn, kinetic_energy_fn = make_loss(network, K, rho0, args.N, dx, args.max_iter, args.epsilon)

def loss_fn(params, vs, rhos, Es, mus): 
    total_loss = args.enls * energy_loss(params, vs, rhos, Es, mus) 
    if args.ptls > 0:
        total_loss += args.ptls * potential_loss(params, vs, rhos, Es, mus) 
    if args.dnls > 0:
        total_loss += args.dnls * density_loss(params, vs, rhos, Es, mus) 
    return total_loss

opt_init, opt_update, get_params = optimizers.adam(step_size=args.lr)
opt_state = opt_init(params)

@jax.jit
def training_step(i, key, opt_state):    
    params = get_params(opt_state)
    vs, rhos, Es, mus = get_batch(key, args.batchsize)
    value, grad = jax.value_and_grad(loss_fn)(params, vs, rhos, Es, mus)

    opt_state = opt_update(i, grad, opt_state)
    return value, opt_state

time_of_last_ckpt = time.time()
print('Start training...')
with io.open(save_path+'.log', 'a', buffering=1, newline='\n') as logfile:
    with trange(t_init, args.epochs) as t:
        for epoch in t:
            key, subkey = jax.random.split(key)
            value, opt_state = training_step(epoch, subkey, opt_state)
            t.set_postfix(loss=value)

            message = ('{} ' + 1*'{:.5f} ').format(epoch, value)
            logfile.write(message + u'\n')

            # Checkpointing
            if time.time() - time_of_last_ckpt > 600:
                checkpoint.save(ckpt_save_path, epoch, get_params(opt_state))
                time_of_last_ckpt = time.time()
